import os
import webbrowser

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles

from .routers import solutions

# Initialize default app
app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # type: ignore
    allow_credentials=True,  # type: ignore
    allow_methods=["*"],  # type: ignore
    allow_headers=["*"],  # type: ignore
)


# Initialize api app
api_app = FastAPI()
api_app.include_router(solutions.router)
app.mount("/api", api_app)

# Mount explorer as static files
explorer_path = os.path.join(os.path.dirname(__file__), "explorer")
explorer_url = "/"
app.mount(explorer_url, StaticFiles(directory=explorer_path, html=True), name="explorer")

# Define uvicorn settings
config = uvicorn.Config("main:app", port=8000, log_level="info")
server = uvicorn.Server(config)

if __name__ == "__main__":
    webbrowser.open("http://localhost:8000/", new=2)
    server.run()



import os
from functools import cache
from os import walk
from typing import Any, Optional

import pandas as pd
from fastapi import HTTPException
from zen_garden.postprocess.results import Results  # type: ignore
from zen_temple.utils import get_variable_name

from ..config import config
from ..models.solution import (
    DataResult,
    SolutionDetail,
    SolutionList,
)


class SolutionRepository:
    def get_list(self) -> list[SolutionList]:
        """
        Creates a list of Solution-objects of all solutions that are contained in any folder contained in the configured SOLUTION_FOLDER.

        This function is very forgiving, it tries to instanciate a Solution for all folders in SOLUTION_FOLDER that contain a 'scenarios.json' file.
        If this fails, it skips the folder.
        """
        solutions_folders: set[str] = set()
        ans = []
        for dirpath, _, filenames in walk(config.SOLUTION_FOLDER):
            if "scenarios.json" in filenames:
                solutions_folders.add(dirpath)

        for folder in solutions_folders:
            try:
                ans.append(SolutionList.from_path(folder))
            except (FileNotFoundError, NotADirectoryError):
                continue
        return ans

    @cache
    def get_detail(self, solution_name: str) -> SolutionDetail:
        """
        Returns the SolutionDetail of a solution given its name.

        The solution name can contain dots which are treated as folders.
        So for example foo/bar.solution will resolve to the solition contained in foo/bar/solution, relative to
        the SOLUTION_FOLDER config value.

        :param solution_name: Name of the solution
        """
        path = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        return SolutionDetail.from_path(path)

    @cache
    def get_full_ts(
        self,
        solution_name: str,
        component: str,
        scenario: Optional[str] = None,
        year: Optional[int] = None,
    ) -> DataResult:
        """
        Returns the full ts and the unit of a component given the solution name, the component name and the scenario name.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :param component: Name of the component.
        :param scenario: Name of the scenario. If skipped, the first scenario is taken.
        :param year: The year of the ts. If skipped, the first year is taken.
        """
        solution_folder = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        unit = self.get_unit(solution_name, component)
        results = Results(solution_folder)

        if year is None:
            year = 0

        full_ts = results.get_full_ts(component, scenario_name=scenario, year=year)
        full_ts = full_ts[~full_ts.index.duplicated(keep="first")]
        full_ts = full_ts.loc[~(full_ts == 0).all(axis=1)]

        return DataResult(data_csv=str(full_ts.to_csv()), unit=unit)

    @cache
    def get_total(
        self, solution_name: str, component: str, scenario: Optional[str] = None
    ) -> DataResult:
        """
        Returns the total and the unit of a component given the solution name, the scenario name and the component name.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :param component: Name of the component.
        :param scenario: Name of the scenario. If skipped, the first scenario is taken.
        """
        solution_folder = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        results = Results(solution_folder)
        unit = self.get_unit(solution_name, component)
        try:
            total: pd.DataFrame | pd.Series[Any] = results.get_total(
                component, scenario_name=scenario
            )
        except KeyError:
            raise HTTPException(status_code=404, detail=f"{component} not found!")

        if type(total) is not pd.Series:
            total = total.loc[~(total == 0).all(axis=1)]  # type: ignore

        return DataResult(data_csv=str(total.to_csv()), unit=unit)

    def get_unit(self, solution_name: str, component: str) -> Optional[str]:
        """
        Returns the unit of a component given the solution name. If there are several units in the requested component, it returns it in form of a CSV string.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        """
        solution_folder = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        results = Results(solution_folder)

        unit_str: str | None = None
        try:
            unit: str | pd.DataFrame = results.get_unit(component)
            if type(unit) is str:
                unit = pd.DataFrame({0: [unit]})
            unit_str = str(unit.to_csv())  # type: ignore

        except Exception:
            unit_str = None
        return unit_str

    @cache
    def get_energy_balance(
        self,
        solution_name: str,
        node: str,
        carrier: str,
        scenario: Optional[str] = None,
        year: Optional[int] = None,
    ) -> dict[str, str]:
        """
        Returns the energy balance dataframes of a solution.
        It drops duplicates of all dataframes and removes the variables that only contain zeros.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :param node: The name of the node.
        :param carrier: The name of the carrier.
        :param scenario: The name of the scenario. If skipped, the first scenario is taken.
        :param year: The desired year. If skipped, the first year is taken.
        """
        solution_folder = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        results = Results(solution_folder)

        if year is None:
            year = 0

        balances: dict[str, pd.DataFrame | pd.Series[Any]] = (
            results.get_energy_balance_dataframes(node, carrier, year, scenario)
        )

        # Drop duplicates of all dataframes
        balances = {
            key: val[~val.index.duplicated(keep="first")] for key, val in balances.items()
        }

        # Drop variables that only contain zeros (except for demand)
        for key, series in balances.items():
            demand_name = get_variable_name(
                "demand", results.get_analysis().zen_garden_version
            )
            if key == demand_name:
                continue

            if type(series) is not pd.Series:
                balances[key] = series.loc[~(series == 0).all(axis=1)]  # type: ignore

        ans = {key: val.to_csv() for key, val in balances.items()}

        return ans


solution_repository = SolutionRepository()



from typing import Optional

from fastapi import APIRouter

from ..models.solution import DataResult, SolutionDetail, SolutionList
from ..repositories.solution_repository import solution_repository

router = APIRouter(prefix="/solutions", tags=["Solutions"])


@router.get("/list")
async def get_list() -> list[SolutionList]:
    """
    Get a list of the available solutions.
    """
    return solution_repository.get_list()


@router.get("/get_detail/{solution_name}")
async def get_detail(solution_name: str) -> SolutionDetail:
    """
    Get the details of a solution.
    """
    ans = solution_repository.get_detail(solution_name)
    return ans


@router.get("/get_total/{solution_name}/{variable_name}")
async def get_total(
    solution_name: str, variable_name: str, scenario: Optional[str] = None
) -> DataResult:
    """
    Get the total of a variable given the solution name, the variable name, and the scenario. If no scenario is provided, the first scenarios in the list is taken.
    """
    ans = solution_repository.get_total(solution_name, variable_name, scenario)
    return ans


@router.get("/get_full_ts/{solution_name}/{variable_name}")
async def get_full_ts(
    solution_name: str,
    variable_name: str,
    scenario: Optional[str] = None,
    year: Optional[int] = None,
) -> DataResult:
    """
    Get the total of a variable given the solution name, the variable name, and the scenario. If no scenario is provided, the first scenarios in the list is taken.
    """
    ans = solution_repository.get_full_ts(solution_name, variable_name, scenario, year)
    return ans


@router.get("/get_unit/{solution_name}/{variable_name}")
async def get_unit(solution_name: str, variable_name: str) -> Optional[str]:
    """
    Get the unit of a variable given the solution name, the variable name, and the scenario. If no scenario is provided, the first scenarios in the list is taken.
    """
    ans = solution_repository.get_unit(solution_name, variable_name)
    return ans


@router.get("/get_energy_balance/{solution_name}/{node_name}/{carrier_name}")
async def get_energy_balance(
    solution_name: str,
    node_name: str,
    carrier_name: str,
    scenario: Optional[str] = None,
    year: Optional[int] = 0,
) -> dict[str, str]:
    """
    Get the energy balance of a specific node and carrier given the solution name, the node name, the carrier, the scenario, and the year.
    If no scenario and/or year is provided, the first one is taken.
    """
    ans = solution_repository.get_energy_balance(
        solution_name, node_name, carrier_name, scenario, year
    )
    return ans



import json
import os
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel, Field
from zen_garden.model.default_config import System  # type: ignore
from zen_garden.postprocess.results import Results  # type: ignore

from zen_temple.utils import get_variable_name

from ..config import config


class ScenarioDetail(BaseModel):
    """
    ScenarioDetail is the model that includes all the detail information of a scenario. It also contains the System-information from ZEN Garden.
    """

    system: System
    reference_carrier: dict[str, str]
    carriers_import: list[str]
    carriers_export: list[str]
    carriers_input: dict[str, list[str]]
    carriers_output: dict[str, list[str]]
    carriers_demand: list[str]
    edges: dict[str, str]


class SolutionDetail(BaseModel):
    """
    SolutionDetail is the model that includes all the detail information of a solution. This includes the ScenarioDetail for all scenarios of a solution.
    """

    name: str
    folder_name: str
    scenarios: dict[str, ScenarioDetail]
    components: list[str]
    version: str

    @staticmethod
    def from_path(path: str) -> "SolutionDetail":
        """
        Generator that instanciates a SolutionDetail given the path of a solution.
        It creates a Solution-instance of ZEN Gardens soluion class and extracts the necessary dataframes from this soultion.

        :param path: Path to the results folder.
        """
        name = os.path.split(path)[-1]
        relative_path = os.path.relpath(path, start=config.SOLUTION_FOLDER)
        results = Results(path)
        results_version = results.get_analysis().zen_garden_version

        reference_carriers = results.get_df(
            get_variable_name("set_reference_carriers", results_version)
        )

        scenario_details = {}

        for scenario_name, scenario in results.solution_loader.scenarios.items():
            system = scenario.system
            reference_carrier = reference_carriers[scenario_name].to_dict()

            df_import = results.get_df(
                get_variable_name("availability_import", results_version)
            )[scenario_name]

            df_export = results.get_df(
                get_variable_name("availability_export", results_version)
            )[scenario_name]

            df_demand = results.get_df(get_variable_name("demand", results_version))[
                scenario_name
            ]

            df_input_carriers = results.get_df(
                get_variable_name("set_input_carriers", results_version)
            )[scenario_name]

            df_output_carriers = results.get_df(
                get_variable_name("set_output_carriers", results_version)
            )[scenario_name]

            edges = results.get_df(
                get_variable_name("set_nodes_on_edges", results_version)
            )[scenario_name]

            edges_dict = edges.to_dict()
            carriers_input_dict = {
                key: val.split(",") for key, val in df_input_carriers.to_dict().items()
            }
            carriers_output_dict = {
                key: val.split(",") for key, val in df_output_carriers.to_dict().items()
            }

            for key in carriers_output_dict:
                if carriers_output_dict[key] == [""]:
                    carriers_output_dict[key] = []

            for key in carriers_input_dict:
                if carriers_input_dict[key] == [""]:
                    carriers_input_dict[key] = []

            carriers_import = list(
                df_import.loc[df_import != 0].index.get_level_values("carrier").unique()
            )
            carriers_export = list(
                df_export.loc[df_export != 0].index.get_level_values("carrier").unique()
            )

            carriers_demand = list(
                df_demand.loc[df_demand != 0].index.get_level_values("carrier").unique()
            )

            scenario_details[scenario_name] = ScenarioDetail(
                system=system,
                reference_carrier=reference_carrier,
                carriers_import=carriers_import,
                carriers_export=carriers_export,
                carriers_input=carriers_input_dict,
                carriers_output=carriers_output_dict,
                carriers_demand=carriers_demand,
                edges=edges_dict,
            )

        version = results.get_analysis().zen_garden_version
        if version is None:
            version = "0.0.0"

        return SolutionDetail(
            name=name,
            folder_name=str(relative_path),
            scenarios=scenario_details,
            components=results.solution_loader.components.keys(),
            version=version,
        )


class SolutionList(BaseModel):
    """
    SolutionList defines the model of the data that is included in the solutions list endpoint.
    """

    folder_name: str
    name: str
    nodes: list[str] = Field(default=[])
    total_hours_per_year: int
    optimized_years: int
    technologies: list[str] = Field(default=[])
    carriers: list[str] = Field(default=[])
    scenarios: list[str] = Field(default=[])

    @staticmethod
    def from_path(path: str) -> "SolutionList":
        """
        Generator method to instantiate a SolutionList ins given the path of a solution.

        :param path: Path to the results folder.
        """
        with open(os.path.join(path, "scenarios.json"), "r") as f:
            scenarios_json: dict[str, Any] = json.load(f)

        scenarios = list(scenarios_json.keys())

        scenario_name = ""

        if len(scenarios_json) > 1:
            first_scenario_name = scenarios[0]
            scenario_name = (
                "scenario_" + scenarios_json[first_scenario_name]["sub_folder"]
            )

        with open(os.path.join(path, scenario_name, "system.json")) as f:
            system: dict[str, Any] = json.load(f)

        relative_folder = path.replace(config.SOLUTION_FOLDER, "")

        if relative_folder[0] == "/":
            relative_folder = relative_folder[1:]

        system["folder_name"] = relative_folder

        system["carriers"] = system["set_carriers"]
        system["technologies"] = system["set_technologies"]
        system["scenarios"] = scenarios
        system["nodes"] = system["set_nodes"]

        scenario_path = Path(path).relative_to(config.SOLUTION_FOLDER)
        system["name"] = ".".join(scenario_path.parts)
        solution = SolutionList(**system)

        return solution


class DataResult(BaseModel):
    """
    DataResult defines the default model that is used to return the CSV string of a dataframe.
    """

    data_csv: str
    unit: Optional[str]
